/*
 * SNIA DDF1 v1.0 metadata format handler.
 *
 * Copyright (C) 2005-2006 IBM, All rights reserved.
 * Written by Darrick Wong <djwong@us.ibm.com>
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include <errno.h>
#include <netinet/in.h>

#define	HANDLER	"ddf1"

#include "internal.h"
#define	FORMAT_HANDLER
#include "ddf1.h"

#define DM_BYTEORDER_SWAB
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/* Number of config records */
#define NUM_CONFIG_ENTRIES(ddf1) ((ddf1)->primary->config_record_len / \
	(ddf1)->primary->vd_config_record_len)

/* Print DDF GUIDs. */
#ifdef  NATIVE_LOG_OFFSET

#define DP_BUF(name, basevar, x, len) do { \
	_dp_guid(lc, name, P_OFF(x, basevar, x), len);\
} while (0)

#define DP_GUID(name, basevar, x) do {\
_dp_guid(lc, name, P_OFF(x, basevar, x), DDF1_GUID_LENGTH);\
} while (0)

static void _dp_guid(struct lib_context *lc, const char *name,
		    unsigned int offset, void *data, unsigned int len)
{
	char *p;
	int i;

	p = data;
	log_print_nnl(lc, "0x%03x %s\"", offset, name);

	for (i = 0; i < len; i++)
		log_print_nnl(lc, "%c",
			      (isgraph(p[i]) || p[i] == ' ' ? p[i] : '.'));

	log_print_nnl(lc, "\" [");

	for (i = 0; i < len; i++)
		log_print_nnl(lc, "%s%02x", (i != 0 ? " " : ""), p[i] & 0xFF);

	log_print_nnl(lc, "]\n");
}

#else
#define DP_BUF(name, basevar, x, len)
#define DP_GUID(name, basevar, x)
#endif

/* Macros to access config records */
#define SR(ddf, idx)	((struct ddf1_spare_header *)(((uint8_t *)(ddf)->cfg) + \
	((idx) * (ddf)->primary->vd_config_record_len * 512)))

#define CR(ddf, idx)	((struct ddf1_config_record *)(((uint8_t *)(ddf)->cfg) + \
	((idx) * (ddf)->primary->vd_config_record_len * 512)))

#define CR_IDS(ddf, cr)	((uint32_t *)(((uint8_t *)(cr)) + \
	sizeof(struct ddf1_config_record)))

#define CR_OFF(ddf, cr)	((uint64_t *)(((uint8_t *)(cr)) + \
	sizeof(struct ddf1_config_record) + \
	(cr_off_maxpds_helper(ddf) * sizeof(uint32_t))))

/* PCI IDs for Adaptec */
#define PCI_VENDOR_ID_ADAPTEC		0x9004
#define PCI_VENDOR_ID_ADAPTEC2		0x9005

/* Helper for CR_OFF */
static uint16_t cr_off_maxpds_helper(struct ddf1 *ddf1)
{
	/* The 0xFFFF nonsense is a weird Adaptec quirk */
	return ddf1->primary->max_primary_elements == 0xFFFF &&
	       ddf1->adaptec_mode ?
		ddf1->primary->max_phys_drives :
		ddf1->primary->max_primary_elements;
}

/* Find the beginning of all DDF metadata */
static uint64_t find_ddf_beginning(struct ddf1 *d)
{
	uint64_t ret = d->anchor_offset;

	if (d->anchor.primary_table_lba < ret)
		ret = d->anchor.primary_table_lba;

	if (d->anchor.secondary_table_lba < ret)
		ret = d->anchor.secondary_table_lba;

#ifdef WORKSPACE_IS_PART_OF_DDF
	if (d->primary->workspace_lba < ret)
		ret = d->primary->workspace_lba;
#endif

	return ret;
}

/* Map DDF1 disk status to dmraid status */
static enum status disk_status(struct ddf1_phys_drive *disk) {
	struct {
		uint8_t	flag;
		enum status status;
	} states[] = {
		{ 0x72, s_broken }, /* 0x40, 0x20, 0x10 and 0x02 */
		{ 0x04, s_nosync },
		{ 0x08, s_setup },
		{ 0x01, s_ok },
	}, *s = states;

	if (disk) {
		do {
			if (disk->state & s->flag)
				return s->status;
		} while ((s++)->status != s_ok);
	}

	return s_undef;
}

/* Find the physical drive data for a drive */
static struct ddf1_phys_drive *get_phys_drive(struct ddf1 *ddf1, uint32_t ref)
{
	unsigned int i = ddf1->pd_header->max_drives;
	struct ddf1_phys_drive *pds;

	while (i--) {
		pds = ddf1->pds + i;
		if (pds->reference == ref)
			return pds;
	}

	return NULL;
}

/*
 * Compare two GUIDs.  For some reason, Adaptec sometimes writes 0xFFFFFFFF
 * as the last four bytes (ala DDF2) and sometimes writes real data.
 * For now we'll compare the first twenty and only the last four if
 * both GUIDs don't have 0xFFFFFFFF in bytes 20-23.  Gross.
 */
static inline uint8_t _and(uint8_t *p)
{
	return p[20] & p[21] & p[22] & p[23];
}

static int guidcmp(uint8_t *one, uint8_t *two)
{
	int x = memcmp(one, two, DDF1_GUID_LENGTH - 4);

	if (x)
		return x;

	return (_and(one) || _and(two)) ? 0 : memcmp(one + 20, two + 20, 4);
}

/* Find the virtual drive that goes with this config record */
static struct ddf1_virt_drive *get_virt_drive(struct ddf1 *ddf1,
					      struct ddf1_config_record *cr)
{
	unsigned int i = ddf1->vd_header->num_drives;
	struct ddf1_virt_drive *vd;

	while (i--) {
		vd = ddf1->vds + i;
		if (!guidcmp(vd->guid, cr->guid))
			return vd;
	}

	return NULL;
}

/* Find this drive's physical data */
static inline struct ddf1_phys_drive *get_this_phys_drive(struct ddf1 *ddf1)
{
	return get_phys_drive(ddf1, ddf1->disk_data->reference);
}

/*
 * Find the index of the VD config record for this drive.
 * FIXME: Drives can belong to multiple VDs!
 */
static int get_config_index(struct ddf1 *ddf1, struct ddf1_phys_drive *pd)
{
	int i, j;
	uint32_t *ids;
	struct ddf1_config_record *cr;

	for (i = 0; i < NUM_CONFIG_ENTRIES(ddf1); i++) {
		cr = CR(ddf1, i);
		if (cr->signature != DDF1_VD_CONFIG_REC)
			continue;

		ids = CR_IDS(ddf1, cr);
		for (j = 0; j < cr->primary_element_count; j++) {
			if (ids[j] == pd->reference)
				return i;
		}
	}

	return -ENOENT;
}

/*
 * Find the VD config record for this drive.
 * FIXME: Drives can belong to multiple VDs!
 */
static inline struct ddf1_config_record *get_config(struct ddf1 *ddf1,
						    struct ddf1_phys_drive *pd)
{
	int x = get_config_index(ddf1, pd);

	return x < 0 ? NULL : CR(ddf1, x);
}
 
/* Find this drive's config record */
static inline struct ddf1_config_record *get_this_config(struct ddf1 *ddf1)
{
	return get_config(ddf1, get_this_phys_drive(ddf1));
}

/* Find the config record disk/offset entry for this config/drive. */
static int get_offset_entry(struct ddf1 *ddf1, struct ddf1_config_record *cr,
			    struct ddf1_phys_drive *pd)
{
	int i;
	uint32_t *ids;

	if (cr) {
		ids = CR_IDS(ddf1, cr);
		for (i = 0; i < ddf1->primary->max_phys_drives; i++) {
			if (ids[i] == pd->reference)
				return i;
		}
	}

	return -ENOENT;
}

/* Find the offset for this config/drive. */
static uint64_t get_offset(struct ddf1 *ddf1, struct ddf1_config_record *cr,
			   struct ddf1_phys_drive *pd)
{
	int x;

	x = get_offset_entry(ddf1, cr, pd);
	if (x > -1)
		return CR_OFF(ddf1, cr)[x];

	return pd->size;
}

/* Calculate the stripe size, in sectors */
static inline unsigned int stride(struct ddf1_config_record *cr)
{
	return 1 << cr->stripe_size; // * 512 * 2;
}

/* Mapping of template types to generic types */
/* FIXME: Does dm do RAID3/4/5 yet? */
static struct types types[] = {
	{ DDF1_RAID0,	t_raid0 },
	{ DDF1_RAID1,	t_raid1 },
/*	{ DDF1_RAID3,	t_raid3 }, */
	{ DDF1_RAID4,	t_raid4 },
	{ DDF1_RAID6,	t_raid6 },
	{ DDF1_CONCAT,	t_linear },
	{ 0, t_undef}
};

/* Seperate array for RAID5 qualifiers */
static struct types qualifier_types[] = {
	/* FIXME: Is RLQ=0 really right symmetric? */
	{ DDF1_RAID5_RS, t_raid5_rs },
	{ DDF1_RAID5_LA, t_raid5_la },
	{ DDF1_RAID5_LS, t_raid5_ls },
	{ 0, t_undef}
};

/* Map the DDF1 raid type codes into dmraid type codes. */
static enum type type(struct lib_context *lc, struct ddf1 *ddf1,
		      struct ddf1_config_record *cr)
{
	unsigned int l;
	struct types *t;

	if (!cr)
		return t_undef;

	l = cr->raid_level;
	if (l == DDF1_JBOD && ddf1->adaptec_mode)
		return t_linear;

	if (l == DDF1_RAID5) {
		/*
		 * FIXME: Do _all_ Adaptec controllers use left
		 * asymmetric parity and write zero to RLQ?
		 */
		if (ddf1->adaptec_mode)
			return t_raid5_la;

		l = cr->raid_qualifier;
		t = qualifier_types;
	} else
		t = types;

	return rd_type(t, l);
}

/* Figure out what endian conversions we need */
static void find_endian(struct lib_context *lc, struct ddf1 *ddf1)
{
	uint8_t *ptr = (uint8_t *)&ddf1->anchor.signature;

	if (ptr[0] == 0xDE && ptr[1] == 0x11)
		ddf1->disk_format = BIG_ENDIAN;
	else if (ptr[0] == 0x11 && ptr[1] == 0xDE)
		ddf1->disk_format = LITTLE_ENDIAN;
	else {
		log_err(lc, "Can't figure out endianness!");
		ddf1->disk_format = 0;
	}
}

/* Convert a DDF header */
static void cvt_header(struct ddf1 *ddf1, struct ddf1_header *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->signature);
	CVT32(hdr->crc);
	CVT32(hdr->seqnum);
	CVT32(hdr->timestamp);
	CVT64(hdr->primary_table_lba);
	CVT64(hdr->secondary_table_lba);
	CVT32(hdr->workspace_length);
	CVT64(hdr->workspace_lba);
	CVT16(hdr->max_phys_drives);
	CVT16(hdr->max_virt_drives);
	CVT16(hdr->max_partitions);
	CVT16(hdr->vd_config_record_len);
	CVT16(hdr->max_primary_elements);
	CVT32(hdr->adapter_data_offset);
	CVT32(hdr->adapter_data_len);
	CVT32(hdr->phys_drive_offset);
	CVT32(hdr->phys_drive_len);
	CVT32(hdr->virt_drive_offset);
	CVT32(hdr->virt_drive_len);
	CVT32(hdr->config_record_offset);
	CVT32(hdr->config_record_len);
	CVT32(hdr->disk_data_offset);
	CVT32(hdr->disk_data_len);
	CVT32(hdr->badblock_offset);
	CVT32(hdr->badblock_len);
	CVT32(hdr->diag_offset);
	CVT32(hdr->diag_len);
	CVT32(hdr->vendor_offset);
	CVT32(hdr->vendor_len);
}

/* Convert DDF adapter data */
static void cvt_adapter(struct ddf1 *ddf1, struct ddf1_adapter *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->signature);
	CVT32(hdr->crc);
	CVT16(hdr->pci_vendor);
	CVT16(hdr->pci_device);
	CVT16(hdr->pci_subvendor);
	CVT16(hdr->pci_subdevice);
}

/* Convert physical disk data */
static void cvt_disk_data(struct ddf1 *ddf1, struct ddf1_disk_data *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->signature);
	CVT32(hdr->crc);
	CVT32(hdr->reference);
}

/* Convert physical drive header data */
static void cvt_phys_drive_header(struct ddf1 *ddf1,
				  struct ddf1_phys_drives *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->signature);
	CVT32(hdr->crc);
	CVT16(hdr->num_drives);
	CVT16(hdr->max_drives);
}

/* Convert physical drive data */
static void cvt_phys_drive(struct ddf1 *ddf1, struct ddf1_phys_drive *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->reference);
	CVT16(hdr->type);
	CVT16(hdr->state);
	CVT64(hdr->size);
}

/* Convert virtual drive header data */
static void cvt_virt_drive_header(struct ddf1 *ddf1,
				  struct ddf1_virt_drives *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->signature);
	CVT32(hdr->crc);
	CVT16(hdr->num_drives);
	CVT16(hdr->max_drives);
}

/* Convert virtual drive data */
static void cvt_virt_drive(struct ddf1 *ddf1, struct ddf1_virt_drive *hdr)
{
	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(hdr->vd_num);
	CVT32(hdr->type);
}

/* Convert config record data */
static void cvt_config_record(struct ddf1 *ddf1,
			      struct ddf1_config_record *hdr)
{
	uint32_t *ids;
	uint64_t *off;
	uint16_t max_pds, i;
	uint32_t x;

	if (BYTE_ORDER == ddf1->disk_format)
		return;

	max_pds = hdr->primary_element_count;
	ids = CR_IDS(ddf1, hdr);

	/* This chunk is derived from CR_OFF */
	x = cr_off_maxpds_helper(ddf1);

	if (ddf1->primary->signature == DDF1_HEADER_BACKWARDS)
		CVT32(x);

	off = ((uint64_t *)(((uint8_t *)hdr) +
		sizeof(struct ddf1_config_record) + (x * sizeof(uint32_t))));

	CVT32(hdr->signature);
	CVT32(hdr->crc);
	CVT32(hdr->timestamp);
	CVT32(hdr->seqnum);
	CVT16(hdr->primary_element_count);

	if (!ddf1->in_cpu_format)
		max_pds = hdr->primary_element_count;

	CVT64(hdr->sectors);
	CVT64(hdr->size);
	CVT32(hdr->spares[0]);
	CVT32(hdr->spares[1]);
	CVT32(hdr->spares[2]);
	CVT32(hdr->spares[3]);
	CVT32(hdr->spares[4]);
	CVT32(hdr->spares[5]);
	CVT32(hdr->spares[6]);
	CVT32(hdr->spares[7]);
	CVT64(hdr->cache_policy);

	for (i = 0; i < max_pds; i++) {
		CVT32(ids[i]);
		CVT64(off[i]);
	}
}

/* Convert spare records */
static void cvt_spare_record(struct ddf1 *ddf1, struct ddf1_spare_header *sh)
{
	uint16_t x, i;

	if (BYTE_ORDER == ddf1->disk_format)
		return;

	CVT32(sh->signature);
	CVT32(sh->crc);
	CVT32(sh->timestamp);
	CVT16(sh->max_spares);
	x = sh->num_spares;
	CVT16(sh->num_spares);
	if (!ddf1->in_cpu_format)
		x = sh->num_spares;

	for (i = 0; i < x; i++)
		CVT16(sh->spares[i].secondary_element);
}

/* Convert endianness of all metadata */
static void cvt_all(struct ddf1 *ddf1)
{
	uint16_t pds, vds;
	uint32_t cfgs, x;
	int i;

	vds = pds = 0;
	cfgs = 0;

	cvt_header(ddf1, &ddf1->anchor);
	if (ddf1->in_cpu_format) {
		cfgs = NUM_CONFIG_ENTRIES(ddf1);

		for (i = 0; i < cfgs; i++) {
			x = *((uint32_t *)CR(ddf1, i));
			if (!ddf1->in_cpu_format &&
			    BYTE_ORDER != ddf1->disk_format)
				CVT32(x);

			switch (x) {
				case DDF1_VD_CONFIG_REC:
					cvt_config_record(ddf1, CR(ddf1, i));
					break;

				case DDF1_SPARE_REC:
					cvt_spare_record(ddf1, SR(ddf1, i));
					break;

				case 0: /* Adaptec controllers set this to 0 */
				case DDF1_INVALID:
					break;
			}
		}
	}

	cvt_header(ddf1, ddf1->primary);
	if (!ddf1->in_cpu_format) {
		cfgs = NUM_CONFIG_ENTRIES(ddf1);

		for (i = 0; i < cfgs; i++) {
			x = *((uint32_t *)CR(ddf1, i));
			if (!ddf1->in_cpu_format &&
			    BYTE_ORDER != ddf1->disk_format)
				CVT32(x);

			switch (x) {
				case DDF1_VD_CONFIG_REC:
					cvt_config_record(ddf1, CR(ddf1, i));
					break;

				case DDF1_SPARE_REC:
					cvt_spare_record(ddf1, SR(ddf1, i));
					break;

				case 0: /* Adaptec controllers set this to 0 */
				case DDF1_INVALID:
					break;
			}
		}
	}

	if (ddf1->secondary)
		cvt_header(ddf1, ddf1->secondary);

	if (ddf1->adapter)
		cvt_adapter(ddf1, ddf1->adapter);

	cvt_disk_data(ddf1, ddf1->disk_data);

	if (ddf1->in_cpu_format)
		pds = ddf1->pd_header->num_drives;

	cvt_phys_drive_header(ddf1, ddf1->pd_header);

	if (!ddf1->in_cpu_format)
		pds = ddf1->pd_header->num_drives;

	for (i = 0; i < pds; i++)
		cvt_phys_drive(ddf1, &ddf1->pds[i]);

	if (ddf1->in_cpu_format)
		vds = ddf1->vd_header->num_drives;

	cvt_virt_drive_header(ddf1, ddf1->vd_header);

	if (!ddf1->in_cpu_format)
		vds = ddf1->vd_header->num_drives;

	for (i = 0; i < vds; i++)
		cvt_virt_drive(ddf1, &ddf1->vds[i]);

	ddf1->in_cpu_format = !ddf1->in_cpu_format;
}

/*
 * Read an DDF1 RAID device.  Fields are little endian, so
 * need to convert them if we're on a BE machine (ppc, etc).
 */
static inline void cond_free(void *p)
{
	if (p)
		dbg_free(p);
}

static int read_extended(struct lib_context *lc, struct dev_info *di,
			 struct ddf1 *ddf1)
{
	uint64_t where;
	uint32_t x;
	int i;

	/* FIXME: We should verify the checksums... */

	/* Read the primary DDF header */
	if (!(ddf1->primary = alloc_private(lc, handler,
					    sizeof(struct ddf1_header))))
		return 0;

	if (!read_file(lc, handler, di->path, ddf1->primary,
		       sizeof(struct ddf1_header),
		       ddf1->anchor.primary_table_lba * DDF1_BLKSIZE))
		goto bad;

	cvt_header(ddf1, ddf1->primary);

	if (ddf1->primary->signature != DDF1_HEADER) {
		log_warn(lc, "Incorrect primary header signature %x.\n",
			 ddf1->primary->signature);
		goto bad;
	}

	/* Read the secondary header. */
	/* FIXME: If we encounter an error, we should use the secondary table */
	if (!(ddf1->secondary = alloc_private(lc, handler,
					      sizeof(struct ddf1_header))))
		goto bad;

	if (ddf1->anchor.secondary_table_lba != 0xFFFFFFFFFFFFFFFFULL &&
	    !read_file(lc, handler, di->path, ddf1->secondary,
		       sizeof(struct ddf1_header),
		       ddf1->anchor.secondary_table_lba * DDF1_BLKSIZE)) {
		goto bad;
	}

	cvt_header(ddf1, ddf1->secondary);
	if (ddf1->secondary->signature != DDF1_HEADER) {
		if (ddf1->secondary->signature)
			log_warn(lc, "Bad secondary header signature %x.",
				 ddf1->secondary->signature);
		dbg_free(ddf1->secondary);
		ddf1->secondary = NULL;
	}

	/* Read the adapter data */
	if (!(ddf1->adapter = alloc_private(lc, handler,
					    sizeof(struct ddf1_adapter))))
		goto bad;

	where = ddf1->primary->primary_table_lba +
		ddf1->primary->adapter_data_offset;
	where *= DDF1_BLKSIZE;

	if (ddf1->primary->adapter_data_offset != 0xFFFFFFFF &&
	    !read_file(lc, handler, di->path, ddf1->adapter,
		       sizeof(struct ddf1_adapter),
		       where))
		goto bad;

	cvt_adapter(ddf1, ddf1->adapter);

	if (ddf1->adapter->signature != DDF1_ADAPTER_DATA) {
		if (ddf1->adapter->signature)
			log_warn(lc, "Incorrect adapter data signature %x.",
				 ddf1->adapter->signature);
		free (ddf1->adapter);
		ddf1->adapter = NULL;
	}

	if (ddf1->adapter &&
	    ddf1->adapter->pci_vendor == PCI_VENDOR_ID_ADAPTEC2)
		ddf1->adaptec_mode = 1;

	/* Read physical drive characteristic data */
	if (!(ddf1->disk_data = alloc_private(lc, handler,
					      sizeof(struct ddf1_disk_data))))
		goto bad;

	where = ddf1->primary->primary_table_lba +
		ddf1->primary->disk_data_offset;
	where *= DDF1_BLKSIZE;

	if ((where != di->sectors) &&
	    !read_file(lc, handler, di->path, ddf1->disk_data,
		       sizeof(struct ddf1_disk_data),
		       where))
		goto bad;

	/*
	 * This table isn't technically required, but for now we rely
	 * on it to give us a key into the physical drive table.
	 */
	cvt_disk_data(ddf1, ddf1->disk_data);

	if (ddf1->disk_data->signature != DDF1_FORCED_PD_GUID) {
		log_warn(lc, "Incorrect disk data signature %x.\n",
			 ddf1->disk_data->signature);
		goto bad;
	}

	/* Read physical drive data header */
	if (!(ddf1->pd_header = alloc_private(lc, handler,
					      ddf1->primary->phys_drive_len *
					      DDF1_BLKSIZE)))
		goto bad;

	where = ddf1->primary->primary_table_lba +
		ddf1->primary->phys_drive_offset;
	where *= DDF1_BLKSIZE;

	if (!read_file(lc, handler, di->path, ddf1->pd_header,
			       ddf1->primary->phys_drive_len * DDF1_BLKSIZE,
			       where))
		goto bad;

	cvt_phys_drive_header(ddf1, ddf1->pd_header);
	if (ddf1->pd_header->signature != DDF1_PHYS_DRIVE_REC) {
		log_warn(lc, "%s: Cannot find physical drive records.\n");
		goto bad;
	}

	/* Now read the physical drive data */
	ddf1->pds = (struct ddf1_phys_drive *)(((uint8_t *)ddf1->pd_header) +
		    sizeof (struct ddf1_phys_drives));
	for (i = 0; i < ddf1->pd_header->num_drives; i++) {
		cvt_phys_drive(ddf1, &ddf1->pds[i]);
		/*
		 * Adaptec controllers have a weird bug where this field is
		 * only four bytes ... and the next four are 0xFF.
		 */
		if (ddf1->pds[i].size >> 32 == 0xFFFFFFFF)
			ddf1->pds[i].size &= 0xFFFFFFFF;
	}

	/* Read virtual drive data header */
	if (!(ddf1->vd_header = alloc_private(lc, handler,
					      ddf1->primary->phys_drive_len *
					      DDF1_BLKSIZE)))
		goto bad;

	where = ddf1->primary->primary_table_lba +
		ddf1->primary->virt_drive_offset;
	where *= DDF1_BLKSIZE;

	if (!read_file(lc, handler, di->path, ddf1->vd_header,
			       ddf1->primary->phys_drive_len * DDF1_BLKSIZE,
			       where))
		goto bad;

	cvt_virt_drive_header(ddf1, ddf1->vd_header);
	if (ddf1->vd_header->signature != DDF1_VIRT_DRIVE_REC) {
		log_warn(lc, "%s: Cannot find virtual drive records.\n");
		goto bad;
	}

	/* Now read the virtual drive data */
	ddf1->vds = (struct ddf1_virt_drive *)(((uint8_t *)ddf1->vd_header) +
		    sizeof (struct ddf1_phys_drives));

	for (i = 0; i < ddf1->vd_header->num_drives; i++)
		cvt_virt_drive(ddf1, &ddf1->vds[i]);

	/* Read config data */
	if (!(ddf1->cfg = alloc_private(lc, handler,
					ddf1->primary->config_record_len *
					DDF1_BLKSIZE)))
		goto bad;

	where = ddf1->primary->primary_table_lba +
		ddf1->primary->config_record_offset;
	where *= DDF1_BLKSIZE;

	if (!read_file(lc, handler, di->path, ddf1->cfg,
		       ddf1->primary->config_record_len * DDF1_BLKSIZE, where))
		goto bad;

	/*
	 * Ensure each record is: a config table for VDs; a config table for
	 * spare disks; or vendor-specifc data of some sort.
	 */
	for (i = 0; i < NUM_CONFIG_ENTRIES(ddf1); i++) {
		x = *((uint32_t *)CR(ddf1, i));
		if (BYTE_ORDER != ddf1->disk_format)
			CVT32(x);

		switch (x) {
			case DDF1_VD_CONFIG_REC:
				cvt_config_record(ddf1, CR(ddf1, i));
				break;

			case DDF1_SPARE_REC:
				cvt_spare_record(ddf1, SR(ddf1, i));
				break;

			case 0: /* Adaptec puts zero in this field??? */
			case DDF1_INVALID:
				break;

			default:
				log_warn(lc, "%s: Unknown config record %d.",
					 di->path, x);
		}
	}

	/*
	 * FIXME: We don't pick up diagnostic logs, vendor specific logs,
	 * bad block data, etc.  That shouldn't cause a problem with reading
	 * or writing metadata, but at some point we might want to do something
	 * with them.
	 */
	ddf1->in_cpu_format = 1;

	return 1;

bad:
	ddf1->vds = NULL;
	ddf1->pds = NULL;
	cond_free(ddf1->cfg);
	cond_free(ddf1->pd_header);
	cond_free(ddf1->disk_data);
	cond_free(ddf1->adapter);
	cond_free(ddf1->secondary);
	cond_free(ddf1->primary);

	return 0;
}

/* Is this DDF1 metadata? */
static int is_ddf1(struct lib_context *lc, struct dev_info *di,
		   struct ddf1 *ddf1)
{
	/*
	 * Check our magic numbers and that the version == v2.
	 * We don't support anything other than that right now.
	 */

	/* FIXME: We should examine the version headers... */
	return ddf1->anchor.signature == DDF1_HEADER ||
	       ddf1->anchor.signature == DDF1_HEADER_BACKWARDS;
}

/* Try to find DDF1 metadata at a given offset (ddf1_sboffset) */
static struct ddf1 *try_to_find_ddf1(struct lib_context *lc,
				     struct dev_info *di,
				     size_t *sz, uint64_t *offset,
				     union read_info *info,
				     uint64_t ddf1_sboffset)
{
	struct ddf1 *ddf1;

	/*
	 * Try to find a DDF1 anchor block at ddf1_sboffset.  In theory this
	 * should be the very last block, but some Adaptec controllers have
	 * issues with standards compliance.  So we have to try with various
	 * offsets.
	 */
	if (!(ddf1 = alloc_private(lc, handler, sizeof(struct ddf1))))
		goto out;

	if (!read_file(lc, handler, di->path, &ddf1->anchor, DDF1_BLKSIZE,
		       ddf1_sboffset))
		goto bad;

	if (!is_ddf1(lc, di, ddf1))
		goto bad;

	ddf1->anchor_offset = ddf1_sboffset;

	/* Convert endianness */
	ddf1->in_cpu_format = 0;
	find_endian(lc, ddf1);
	cvt_header(ddf1, &ddf1->anchor);

	/* Read extended metadata. */
	if (read_extended(lc, di, ddf1))
		goto out;

   bad:
	ddf1->anchor_offset = di->sectors;
	dbg_free(ddf1);
	ddf1 = NULL;

   out:
	return ddf1;
}

/*
 * Attempt to interpret DDF1 metadata from a block device.  This function
 * returns either NULL or a pointer to a descriptor struct.
 * Note that the struct should be fully converted to the correct endianness
 * by the time this function returns.
 */
static void *read_metadata_areas(struct lib_context *lc, struct dev_info *di,
				 size_t *sz, uint64_t *offset,
				 union read_info *info)
{
	struct ddf1 *ddf1;

	if ((ddf1 = try_to_find_ddf1(lc, di, sz, offset,
				     info, DDF1_CONFIGOFFSET)))
		goto out;

	if ((ddf1 = try_to_find_ddf1(lc, di, sz, offset, info,
				     DDF1_CONFIGOFFSET_ADAPTEC)))
		ddf1->adaptec_mode = 1;

   out:
	return ddf1;
}

/*
 * Write out parts of the metadata for sample collection.
 * This is all hogwash since file_metadata can only be called once...
 */
static void file_metadata_areas(struct lib_context *lc, struct dev_info *di,
				void *meta)
{
	struct ddf1 *ddf1 = meta;
	uint64_t start = find_ddf_beginning(ddf1);
	uint8_t *buf;

	if (!(buf = dbg_malloc((di->sectors - start) * DDF1_BLKSIZE)))
		LOG_ERR(lc, , "Unable to allocate memory.");

	if (!read_file(lc, handler, di->path, buf,
		       (di->sectors - start) * DDF1_BLKSIZE,
		       start * DDF1_BLKSIZE)) {
		dbg_free(buf);
		LOG_ERR(lc, , "Unable to read metadata.");
	}

	file_metadata(lc, handler, di->path, buf,
		      (di->sectors - start) * DDF1_BLKSIZE,
		      start * DDF1_BLKSIZE);


	/* Record the device size if -D was specified. */
	file_dev_size(lc, handler, di);

	dbg_free(buf);

	return;
}

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct dev_info *di, void *meta, union read_info *info);
static struct raid_dev *ddf1_read(struct lib_context *lc,
					struct dev_info *di)
{
	/*
	 * NOTE: Everything called after read_metadata_areas assumes that
	 * the reserved block, raid table and config table have been
	 * converted to the appropriate endianness.
	 */
	return read_raid_dev(lc, di, read_metadata_areas, 0, 0, NULL, NULL,
			     file_metadata_areas, setup_rd, handler);
}

static inline int compose_id(struct ddf1 *ddf1)
{
	struct ddf1_phys_drive *pd = get_this_phys_drive(ddf1);

	return get_offset_entry(ddf1, get_config(ddf1, pd), pd);
}

/* No sort. */
static int no_sort(struct list_head *pos, struct list_head *new)
{
	return 0;
}

/* Sort DDF1 devices by for a RAID set. */
static int dev_sort(struct list_head *pos, struct list_head *new)
{
	return compose_id(META(RD(new), ddf1)) <
	       compose_id(META(RD(pos), ddf1));
}

#if 0
/*
 * Find the top-level RAID set for an DDF1 context.
 */
static int find_toplevel(struct lib_context *lc, struct ddf1 *ddf1)
{
	int i, toplevel = -1;

	for (i = 0; i < ddf1->rt.elmcnt; i++) {
		if (ddf1->rt.ent[i].raidlevel == FWL)
			toplevel = i;
		else if (ddf1->rt.ent[i].raidlevel == FWL_2) {
			toplevel = i;
			break;
		}
	}
	
	return toplevel;
}

/*
 * Find the logical drive configuration that goes with this
 * physical disk configuration.
 */
static struct ddf1_raid_configline *find_logical(struct ddf1 *ddf1)
{
	int i, j;

	/* This MUST be done backwards! */
	for (i = ddf1->rt.elmcnt - 1; i > -1; i--) {
		if (ddf1->rt.ent[i].raidmagic == ddf1->rb.drivemagic)
		{
			for (j = i - 1; j > -1; j--) {
				if (ddf1->rt.ent[j].raidlevel == FWL)
					return ddf1->rt.ent + j;
			}
		}
	}

	return NULL;
}

/* Wrapper for name() */
static char *js_name(struct lib_context *lc, struct raid_dev *rd,
		     unsigned int subset)
{
	return name(lc, META(rd, ddf1));
}
#endif

/*
 * IO error event handler.
 */
static int event_io(struct lib_context *lc, struct event_io *e_io)
{
	log_err(lc, "I/O error on device %s at sector %lu.\n",
		e_io->rd->di->path, e_io->sector);

	LOG_ERR(lc, 0, "PANIC: ddf1 doesn't know about event_io!\n");
}

#if 0
	struct raid_dev *rd = e_io->rd;
	struct ddf1 *ddf1 = META(rd, ddf1);
	struct ddf1_raid_configline *cl = this_disk(ddf1);
	struct ddf1_raid_configline *fwl = find_logical(ddf1);

	/* Ignore if we've already marked this disk broken(?) */
	if (rd->status & s_broken)
		return 0;
	
	/* Mark the array as degraded and the disk as failed. */
	rd->status = s_broken;
	cl->raidstate = LSU_COMPONENT_STATE_FAILED;
	fwl->raidstate = LSU_COMPONENT_STATE_DEGRADED;
	/* FIXME: Do we have to mark a parent too? */

	/* Indicate that this is indeed a failure. */
	return 1;
}
#endif

#define NAME_SIZE 64
/* Formulate a RAID set name for this disk. */
static char *name(struct lib_context *lc, struct ddf1 *ddf1)
{
	int j;
	char *buf;
	struct ddf1_phys_drive *pd;
	struct ddf1_virt_drive *vd;
	struct ddf1_config_record *cr;

	if (!(buf = dbg_malloc(NAME_SIZE)))
		LOG_ERR(lc, NULL, "Cannot allocate memory for name.");

	if (!(pd = get_this_phys_drive(ddf1)))
		LOG_ERR(lc, NULL, "Cannot find physical drive description!");

 	if (!(cr = get_config(ddf1, pd))) {
		memcpy(buf, ".ddf1_spares", 13);
		return buf;
	}

	if (!(vd = get_virt_drive(ddf1, cr)))
		LOG_ERR(lc, NULL, "Cannot find virtual drive description!");

	memcpy(buf, "ddf1_", 5);

	if (!vd->name[0]) {
		sprintf(buf+5, "%02x%02x%02x%02x", vd->guid[0], vd->guid[1],
			vd->guid[2], vd->guid[3]);
		sprintf(buf+9, "%02x%02x%02x%02x", vd->guid[4], vd->guid[5],
			vd->guid[6], vd->guid[7]);
		sprintf(buf+13, "%02x%02x%02x%02x", vd->guid[8], vd->guid[9],
			vd->guid[10], vd->guid[11]);
		sprintf(buf+17, "%02x%02x%02x%02x", vd->guid[12], vd->guid[13],
			vd->guid[14], vd->guid[15]);
		sprintf(buf+21, "%02x%02x%02x%02x", vd->guid[16], vd->guid[17],
			vd->guid[18], vd->guid[19]);
		sprintf(buf+25, "%02x%02x%02x%02x", vd->guid[20], vd->guid[21],
			vd->guid[22], vd->guid[23]);
		mk_alpha(lc, buf + 5, 48);
	} else
		memcpy(buf+5, vd->name, 16);

	for (j = NAME_SIZE - 1; j >= 0; j--) {
		if (isgraph(buf[j]))
			break;

		buf[j] = 0;
	}

	return buf;
}

/* 
 * Add an DDF1 device to a RAID set.  This involves finding the raid set to
 * which this disk belongs, and then attaching it.  Note that there are other
 * complications, such as two-layer arrays (RAID10).
 *
 * FIXME: We haven't been able to set up a RAID10 for testing...
 */
static struct raid_set *ddf1_group(struct lib_context *lc, struct raid_dev *rd)
{
	int (*rd_sort)();
	struct ddf1 *ddf1 = META(rd, ddf1);
	struct ddf1_phys_drive *pd;
	struct ddf1_config_record *cr;
	struct raid_set *rs;
	char *set_name;

	if (!(pd = get_this_phys_drive(ddf1)))
		LOG_ERR(lc, NULL, "Cannot find physical drive description!\n");

	if (!(set_name = name(lc, ddf1)))
		LOG_ERR(lc, NULL, "%s: Could not find RAID array name.\n",
			rd->di->path);

	if ((rs = find_or_alloc_raid_set(lc, set_name, FIND_TOP, rd,
					 LC_RS(lc), NO_CREATE,
					 NO_CREATE_ARG))) {
		if ((cr = get_config(ddf1, pd))) {
			rs->stride = stride(cr);
			rs->type = type(lc, ddf1, cr);
			rd_sort = dev_sort;
		} else {
			rs->stride = 0;
			rs->type = t_linear;
			rd_sort = no_sort;
		}

		rs->status = s_ok;
		list_add_sorted(lc, &rs->devs, &rd->devs, rd_sort);
	}
	
	return rs;
}

/* Write metadata. */
static int ddf1_write(struct lib_context *lc,  struct raid_dev *rd, int erase)
{
	int ret;
        struct ddf1 *ddf1 = META(rd, ddf1);

        cvt_all(ddf1);
        ret = write_metadata(lc, handler, rd, -1, erase);
        cvt_all(ddf1);

        return ret;
}

/*
 * Check integrity of a RAID set.
 */

/* Retrieve the number of devices that should be in this set. */
static unsigned int device_count(struct raid_dev *rd, void *context)
{
	/* Get the logical drive */
	struct ddf1_config_record *cr = get_this_config(META(rd, ddf1));

	return cr ? cr->primary_element_count : 0;
}

/* Check a RAID device */
static int check_rd(struct lib_context *lc, struct raid_set *rs,
		    struct raid_dev *rd, void *context)
{
	/*
	 * FIXME: Should we do more checking for brokenness here?
	 * We could check SMART data, verify that nobody else is
	 * putting non-ddf disks in our raid set, etc.
	 */
	return rd->type != s_broken;
}

/* Start the recursive RAID set check. */
static int ddf1_check(struct lib_context *lc, struct raid_set *rs)
{
	return check_raid_set(lc, rs, device_count, NULL, check_rd,
			      NULL, handler);
}

static struct event_handlers ddf1_event_handlers = {
	.io = event_io,
	.rd = NULL,	/* FIXME: no device add/remove event handler yet. */
};

#ifdef DMRAID_NATIVE_LOG
/* Dump DDF tables. */
static void dump_header(struct lib_context *lc, struct ddf1_header *dh)
{
	log_print(lc, "DDF Header at %p", dh);
	DP("signature:\t0x%X", dh, dh->signature);
	DP("crc:\t\t0x%X", dh, dh->crc);
	DP_GUID("guid:\t\t", dh, dh->guid);
	DP_BUF("rev:\t\t", dh, dh->ddf_rev, DDF1_REV_LENGTH);
	DP("seqnum:\t\t%d", dh, dh->seqnum);
	DP("timestamp:\t0x%X", dh, dh->timestamp);
	DP("open:\t\t0x%X", dh, dh->open_flag);
	DP("foreign:\t\t0x%X", dh, dh->foreign_flag);
	DP("grouping:\t\t0x%X", dh, dh->grouping_enforced);
	DP("primary header:\t%lu", dh, dh->primary_table_lba);
	DP("secondary header:\t%lu", dh, dh->secondary_table_lba);
	DP("header type:\t0x%X", dh, dh->header_type);
	DP("workspace len:\t%d", dh, dh->workspace_length);
	DP("workspace lba:\t%lu", dh, dh->workspace_lba);
	DP("max pd:\t\t%d", dh, dh->max_phys_drives);
	DP("max vd:\t\t%d", dh, dh->max_virt_drives);
	DP("max part:\t\t%d", dh, dh->max_partitions);
	DP("vd_config len:\t%d", dh, dh->vd_config_record_len);
	DP("max_primary_elts:\t%d", dh, dh->max_primary_elements);
	DP("adapter_offset:\t%d", dh, dh->adapter_data_offset);
	DP("adapter_len:\t%d", dh, dh->adapter_data_len);
	DP("pd_offset:\t%d", dh, dh->phys_drive_offset);
	DP("pd_len:\t\t%d", dh, dh->phys_drive_len);
	DP("vd_offset:\t%d", dh, dh->virt_drive_offset);
	DP("vd_len:\t\t%d", dh, dh->virt_drive_len);
	DP("config_offset:\t%d", dh, dh->config_record_offset);
	DP("config_len:\t%d", dh, dh->config_record_len);
	DP("disk_data_offset:\t%d", dh, dh->disk_data_offset);
	DP("disk_data_len:\t%d", dh, dh->disk_data_len);
	DP("badblock_offset:\t%d", dh, dh->badblock_offset);
	DP("badblock_len:\t%d", dh, dh->badblock_len);
	DP("diag_offset:\t%d", dh, dh->diag_offset);
	DP("diag_len:\t\t%d", dh, dh->diag_len);
	DP("vendor_offset:\t%d", dh, dh->vendor_offset);
	DP("vendor_len:\t%d", dh, dh->vendor_len);
}

static void dump_adapter(struct lib_context *lc, struct ddf1_adapter *da)
{
	log_print(lc, "Adapter Data at %p", da);
	DP("signature:\t0x%X", da, da->signature);
	DP("crc:\t\t0x%X", da, da->crc);
	DP_GUID("guid:\t\t", da, da->guid);
	DP("pci vendor:\t0x%X", da, da->pci_vendor);
	DP("pci device:\t0x%X", da, da->pci_device);
	DP("pci subvendor:\t0x%X", da, da->pci_subvendor);
	DP("pci subdevice:\t0x%X", da, da->pci_subdevice);
}

static void dump_disk_data(struct lib_context *lc, struct ddf1_disk_data *fg)
{
	log_print(lc, "Disk Data at %p", fg);
	DP("signature:\t0x%X", fg, fg->signature);
	DP("crc:\t\t0x%X", fg, fg->crc);
	DP_GUID("guid:\t\t", fg, fg->guid);
	DP("reference:\t\t0x%X", fg, fg->reference);
	DP("forced_ref_flag:\t%d", fg, fg->forced_ref_flag);
	DP("forced_guid_flag:\t%d", fg, fg->forced_guid_flag);
}

static void dump_phys_drive_header(struct lib_context *lc,
				   struct ddf1_phys_drives *pd)
{
	log_print(lc, "Physical Drive Header at %p", pd);
	DP("signature:\t0x%X", pd, pd->signature);
	DP("crc:\t\t0x%X", pd, pd->crc);
	DP("num drives:\t%d", pd, pd->num_drives);
	DP("max drives:\t%d", pd, pd->max_drives);
}

static void dump_phys_drive(struct lib_context *lc, struct ddf1_phys_drive *pd)
{
	log_print(lc, "Physical Drive at %p", pd);
	DP_GUID("guid:\t\t", pd, pd->guid);
	DP("reference #:\t0x%X", pd, pd->reference);
	DP("type:\t\t0x%X", pd, pd->type);
	DP("state:\t\t0x%X", pd, pd->state);
	DP("size:\t\t%llu", pd, pd->size);
	DP_BUF("path info:\t", pd, pd->path_info, 18);
}

static void dump_virt_drive_header(struct lib_context *lc,
				   struct ddf1_virt_drives *vd)
{
	log_print(lc, "Virtual Drive Header at %p", vd);
	DP("signature:\t0x%X", vd, vd->signature);
	DP("crc:\t\t0x%X", vd, vd->crc);
	DP("num drives:\t%d", vd, vd->num_drives);
	DP("max drives:\t%d", vd, vd->max_drives);
}

static void dump_virt_drive(struct lib_context *lc, struct ddf1_virt_drive *vd)
{
	log_print(lc, "Virtual Drive at %p", vd);
	DP_GUID("guid:\t\t", vd, vd->guid);
	DP("vd #:\t\t0x%X", vd, vd->vd_num);
	DP("type:\t\t0x%X", vd, vd->type);
	DP("state:\t\t0x%X", vd, vd->state);
	DP("init state:\t0x%X", vd, vd->init_state);
	DP_BUF("name:\t\t", vd, vd->name, 16);
}

static void dump_config_record(struct lib_context *lc, struct ddf1 *ddf,
			       unsigned int idx)
{
	uint32_t *cfg_drive_ids;
	uint64_t *cfg_drive_offsets;
	struct ddf1_config_record *cfg = CR(ddf, idx);
	uint16_t x;
	int i;

	if (cfg->signature != DDF1_VD_CONFIG_REC)
		return;
	log_print(lc, "Virtual Drive Config Record at %p", cfg);
	DP("signature:\t0x%X", cfg, cfg->signature);
	DP("crc:\t\t0x%X", cfg, cfg->crc);
	DP_GUID("guid:\t\t", cfg, cfg->guid);
	DP("timestamp:\t0x%X", cfg, cfg->timestamp);
	DP("seqnum:\t\t%d", cfg, cfg->seqnum);
	DP("primary count:\t%d", cfg, cfg->primary_element_count);
	DP("stripe size:\t%dKiB", cfg, cfg->stripe_size);
	DP("raid level:\t%d", cfg, cfg->raid_level);
	DP("raid qualifier:\t%d", cfg, cfg->raid_qualifier);
	DP("secondary count:\t%d", cfg, cfg->secondary_element_count);
	DP("secondary number:\t%d", cfg, cfg->secondary_element_number);
	DP("secondary level:\t%d", cfg, cfg->secondary_element_raid_level);
	DP("spare 0:\t\t0x%X", cfg, cfg->spares[0]);
	DP("spare 1:\t\t0x%X", cfg, cfg->spares[1]);
	DP("spare 2:\t\t0x%X", cfg, cfg->spares[2]);
	DP("spare 3:\t\t0x%X", cfg, cfg->spares[3]);
	DP("spare 4:\t\t0x%X", cfg, cfg->spares[4]);
	DP("spare 5:\t\t0x%X", cfg, cfg->spares[5]);
	DP("spare 6:\t\t0x%X", cfg, cfg->spares[6]);
	DP("spare 7:\t\t0x%X", cfg, cfg->spares[7]);
	DP("cache policy:\t0x%X", cfg, cfg->cache_policy);
	DP("bg task rate:\t%d", cfg, cfg->bg_task_rate);
	DP("sector count:\t%llu", cfg, cfg->sectors);
	DP("size:\t\t%llu", cfg, cfg->size);
	cfg_drive_ids = CR_IDS(ddf, cfg);
	cfg_drive_offsets = CR_OFF(ddf, cfg);

	x = cfg->primary_element_count;
	log_print(lc, "Drive map:");
	for (i = 0; i < x; i++) {
		log_print(lc, "%d: %X @ %lu", i, cfg_drive_ids[i],
			  cfg_drive_offsets[i]);
	}
}	

static void dump_spares(struct lib_context *lc, struct ddf1 *ddf1,
			unsigned int idx)
{
	struct ddf1_spare_header *sh;
	int i;

	sh = SR(ddf1, idx);

	log_print(lc, "Spare Config Record at %p", sh);
	DP("signature:\t0x%X", sh, sh->signature);
	DP("crc:\t\t0x%X", sh, sh->crc);
	DP("timestamp:\t0x%X", sh, sh->timestamp);
	DP("type:\t\t0x%X", sh, sh->type);
	DP("num drives:\t%d", sh, sh->num_spares);
	DP("max drives:\t%d", sh, sh->max_spares);

	for (i = 0; i < sh->num_spares; i++) {
		log_print(lc, "Spare %d:", i);
		DP_GUID("guid:\t\t", sh, sh->spares[i].guid);
		DP("secondary:\t%d", sh, sh->spares[i].secondary_element);
	}
}

static void dump_config_records(struct lib_context *lc, struct ddf1 *ddf1)
{
	int i;
	uint32_t x;

	for (i = 0; i < NUM_CONFIG_ENTRIES(ddf1); i++) {
		x = *((uint32_t *)CR(ddf1, i));
		switch (x) {
			case DDF1_VD_CONFIG_REC:
				dump_config_record(lc, ddf1, i);
				break;

			case DDF1_SPARE_REC:
				dump_spares(lc, ddf1, i);
				break;

			case 0: /* Adaptec controllers set this to 0 */
			case DDF1_INVALID:
				break;

			default:
				log_err(lc, "Unknown config record %x", x);
		}
	}
}

/* Dump the entire table */
static void dump_all(struct lib_context *lc, struct ddf1 *ddf1,
		     const char *dev)
{
	int i;
	uint16_t x;

	log_print(lc, "%s (%s):", dev, handler);
	log_print(lc, "DDF1 anchor at %llu.",
		  ddf1->anchor_offset / DDF1_BLKSIZE);
	log_print(lc, "DDF1 tables in %s-endian format.",
		  (ddf1->disk_format == LITTLE_ENDIAN ? "little" : "big"));
	
	dump_header(lc, &ddf1->anchor);
	dump_header(lc, ddf1->primary);
	if (ddf1->secondary)
		dump_header(lc, ddf1->secondary);

	if (ddf1->adapter)
		dump_adapter(lc, ddf1->adapter);

	dump_disk_data(lc, ddf1->disk_data);
	dump_phys_drive_header(lc, ddf1->pd_header);
	x = ddf1->pd_header->num_drives;

	for (i = 0; i < x; i++)
		dump_phys_drive(lc, &ddf1->pds[i]);

	dump_virt_drive_header(lc, ddf1->vd_header);
	x = ddf1->vd_header->num_drives;

	for (i = 0; i < x; i++)
		dump_virt_drive(lc, &ddf1->vds[i]);

	dump_config_records(lc, ddf1);
}

/*
 * Log native information about the RAID device.
 */
static void ddf1_log(struct lib_context *lc, struct raid_dev *rd)
{
	dump_all(lc, META(rd, ddf1), rd->di->path);
}
#endif /* #ifdef DMRAID_NATIVE_LOG  */

static struct dmraid_format ddf1_format = {
	.name	= HANDLER,
	.descr	= "SNIA DDF1",
	.caps	= "0,1,3,4,5,linear",
	.format = FMT_RAID,
	.read	= ddf1_read,
	.write	= ddf1_write,
	.group	= ddf1_group,
	.check	= ddf1_check,
	.events	= &ddf1_event_handlers,
#ifdef DMRAID_NATIVE_LOG
	.log	= ddf1_log,
#endif
};

/* Register this format handler with the format core */
int register_ddf1(struct lib_context *lc)
{
	return register_format_handler(lc, &ddf1_format);
}

/* Figure out the real size of a disk... */
static uint64_t get_size(struct lib_context *lc, struct ddf1 *ddf1,
			 struct ddf1_config_record *cr,
			 struct ddf1_phys_drive *pd)
{
	if (!cr)
		return pd->size;

	switch (type(lc, ddf1, cr)) {
		case t_raid0:
			/* Some Adaptec controllers need this clamping. */
			return cr->sectors - cr->sectors %
			       (1 << cr->stripe_size);

		default:
			return cr->sectors;
	}
}

/*
 * Set up a RAID device from what we've assembled out of the metadata.
 * FIXME: What if we have more than one array per disk?
 */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct dev_info *di, void *meta, union read_info *info)
{
	struct ddf1 *ddf1 = meta;
	struct meta_areas *ma;
	struct ddf1_phys_drive *pd;
	struct ddf1_config_record *cr;
	int md_area_count = 5;

	if (!(pd = get_this_phys_drive(ddf1)))
		LOG_ERR(lc, 0, "Cannot find physical drive description!\n");

	cr = get_config(ddf1, pd);

	/* We need multiple metadata areas */
	if (ddf1->adapter)
		md_area_count++;

	if (ddf1->secondary)
		md_area_count++;

	if (ddf1->disk_data)
		md_area_count++;

	if (ddf1->primary->workspace_lba != 0xFFFFFFFFFFFFFFFFULL)
		md_area_count++;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler,
						md_area_count)))
		return 0;

	ma = rd->meta_areas;
	ma->offset = ddf1->anchor_offset;
	ma->size = DDF1_BLKSIZE;
	ma->area = &ddf1->anchor;

	if (ddf1->secondary) {
		ma++;
		ma->offset = ddf1->primary->secondary_table_lba;
		ma->size = DDF1_BLKSIZE;
		ma->area = ddf1->secondary;
	}

	ma++;
	ma->offset = ddf1->primary->primary_table_lba;
	ma->size = DDF1_BLKSIZE;
	ma->area = ddf1->primary;

	if (ddf1->adapter) {
		ma++;
		ma->offset = ddf1->primary->primary_table_lba +
			     ddf1->primary->adapter_data_offset;
		ma->size = ddf1->primary->adapter_data_len * DDF1_BLKSIZE;
		ma->area = ddf1->adapter;
	}

	if (ddf1->primary->workspace_lba != 0xFFFFFFFFFFFFFFFFULL) {
		ma++;
		ma->offset = ddf1->primary->workspace_lba;
		ma->size = ddf1->primary->workspace_length * DDF1_BLKSIZE;
		ma->area = dbg_malloc(1);
	}

	if (ddf1->disk_data) {
		ma++;
		ma->offset = ddf1->primary->primary_table_lba +
			     ddf1->primary->disk_data_offset;
		ma->size = ddf1->primary->disk_data_len * DDF1_BLKSIZE;
		ma->area = ddf1->disk_data;
	}

	ma++;
	ma->offset = ddf1->primary->primary_table_lba +
		     ddf1->primary->phys_drive_offset;
	ma->size = ddf1->primary->phys_drive_len * DDF1_BLKSIZE;
	ma->area = ddf1->pd_header;

	ma++;
	ma->offset = ddf1->primary->primary_table_lba +
		     ddf1->primary->virt_drive_offset;
	ma->size = ddf1->primary->virt_drive_len * DDF1_BLKSIZE;
	ma->area = ddf1->vd_header;

	ma++;
	ma->offset = ddf1->primary->primary_table_lba +
		     ddf1->primary->config_record_offset;
	ma->size = ddf1->primary->config_record_len * DDF1_BLKSIZE;
	ma->area = ddf1->cfg;

	/* Now set up the rest of the metadata info */
        rd->di = di;
	rd->fmt = &ddf1_format;
	rd->status = disk_status(pd);
	rd->type   = (cr ? type(lc, ddf1, cr) : t_spare);
	rd->offset = get_offset(ddf1, cr, pd);
	rd->sectors = get_size(lc, ddf1, cr, pd);

	return (rd->name = name(lc, ddf1)) ? 1 : 0;
}
